using System;
using System.Data;
using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace NBearMapping.Test
{
    /// <summary>
    /// Summary description for MappingTest
    /// </summary>
    [TestClass]
    public class MappingTest
    {
        private static DataTable table;
        private static User user3 = new User(null, "teddy3", UserStatus.Admin);

        #region Additional test attributes

        //Use ClassInitialize to run code before running the first test in the class
        [ClassInitialize()]
        public static void MyClassInitialize(TestContext testContext)
        {
            table = new DataTable("User");
            table.Columns.Add("UserID", typeof(int));
            table.Columns.Add("Name", typeof(string));
            table.Columns.Add("Status", typeof(int));
            DataRow row = table.NewRow();
            row["UserID"] = 2;
            row["Name"] = "teddy";
            row["Status"] = (int) UserStatus.Admin;
            table.Rows.Add(row);
        }

        #endregion

        [TestMethod]
        public void TestDataRowToObject()
        {
            ObjectMapper mapper = new ObjectMapper(typeof(DataRow), typeof(User));
            mapper.AddCustomMappingName("UserID", "ID");
            User user = (User)mapper.ConvertObject(table.Rows[0]);
            Assert.AreEqual(table.Rows[0]["UserID"], user.ID);
            Assert.AreEqual(table.Rows[0]["Name"], user.Name);
            Assert.AreEqual(table.Rows[0]["Status"], (int)user.Status);
            User user2 = new User();
            mapper.ConvertObject(table.Rows[0], user2);
            Assert.AreEqual(table.Rows[0]["UserID"], user2.ID);
            Assert.AreEqual(table.Rows[0]["Name"], user2.Name);   
            Assert.AreEqual(table.Rows[0]["Status"], (int)user2.Status);
        }

        [TestMethod]
        public void TestDataReaderToObject()
        {
            ObjectMapper mapper = new ObjectMapper(typeof(IDataReader), typeof(User));
            mapper.AddCustomMappingName("UserID", "ID");
            IDataReader reader = table.CreateDataReader();
            reader.Read();
            User user = (User)mapper.ConvertObject(reader);
            Assert.AreEqual(table.Rows[0]["UserID"], user.ID);
            Assert.AreEqual(table.Rows[0]["Name"], user.Name);
            Assert.AreEqual(table.Rows[0]["Status"], (int)user.Status);
            User user2 = new User();
            IDataReader reader2 = table.CreateDataReader();
            reader2.Read();
            mapper.ConvertObject(reader2, user2);
            Assert.AreEqual(table.Rows[0]["UserID"], user2.ID);
            Assert.AreEqual(table.Rows[0]["Name"], user2.Name);
            Assert.AreEqual(table.Rows[0]["Status"], (int)user2.Status);
        }

        [TestMethod]
        public void ObjectToObject()
        {
            ObjectMapper mapper = new ObjectMapper(typeof(User), typeof(User));
            User user = (User)mapper.ConvertObject(user3);
            Assert.AreEqual(user3.ID, user.ID);
            Assert.AreEqual(user3.Name, user.Name);
            Assert.AreEqual(user3.Status, user.Status);
            User user2 = new User();
            mapper.ConvertObject(user3, user2);
            Assert.AreEqual(user3.ID, user2.ID);
            Assert.AreEqual(user3.Name, user2.Name);
            Assert.AreEqual(user3.Status, user2.Status);
        }

        [TestMethod]
        public void TestObjectToDataTableDataReaderAndDataRow()
        {
            ObjectMapper mapper = new ObjectMapper(typeof(User), typeof(DataTable));
            mapper.AddCustomMappingName("ID", "UserID");
            DataTable userTable = (DataTable)mapper.ConvertObject(user3);
            Assert.AreEqual(user3.ID, userTable.Rows[0]["UserID"] == DBNull.Value ? null : userTable.Rows[0]["UserID"]);
            Assert.AreEqual(user3.Name, userTable.Rows[0]["Name"]);
            Assert.AreEqual((int)user3.Status, userTable.Rows[0]["Status"]);
            mapper.ConvertObject(user3, userTable);
            Assert.AreEqual(user3.ID, userTable.Rows[1]["UserID"] == DBNull.Value ? null : userTable.Rows[1]["UserID"]);
            Assert.AreEqual(user3.Name, userTable.Rows[1]["Name"]);
            Assert.AreEqual((int)user3.Status, userTable.Rows[1]["Status"]);

            mapper = new ObjectMapper(typeof(User), typeof(IDataReader));
            mapper.AddCustomMappingName("ID", "UserID");
            IDataReader reader = (IDataReader)mapper.ConvertObject(user3);
            Assert.IsNotNull(reader);

            mapper = new ObjectMapper(typeof(User), typeof(DataRow));
            mapper.AddCustomMappingName("ID", "UserID");
            DataRow row = (DataRow)mapper.ConvertObject(user3);
            Assert.IsNotNull(row);
        }

        [TestMethod]
        public void TestPerformance()
        {
            int rowCount = 100000;

            DataTable userTable = (DataTable)table.Clone();
            for (int i = 0; i < rowCount; ++i)
            {
                userTable.Merge(table, true);
            }

            Console.WriteLine(string.Format("Test performance of converting {0} data rows to class objects", rowCount));
            Console.WriteLine("Manually Coding vs NBearMapping vs Reflection");

            //manually coding
            long ticks = DateTime.Now.Ticks;
            User[] users1 = new User[userTable.Rows.Count];
            for  (int i = 0; i < userTable.Rows.Count; ++i)
            {
                users1[i] = new User((int)userTable.Rows[i]["UserID"], (string)userTable.Rows[i]["Name"], (UserStatus)userTable.Rows[i]["Status"]);
            }
            long ticks1 = DateTime.Now.Ticks - ticks;
            Console.WriteLine(string.Format("{0} (1)", ticks1));
            Console.WriteLine(" vs ");

            //NBearMapping
            ObjectMapper mapper = new ObjectMapper(typeof(DataRow), typeof(User));
            mapper.AddCustomMappingName("UserID", "ID");
            User temp = (User)mapper.ConvertObject(userTable.Rows[0], new User());
            Assert.AreEqual(userTable.Rows[0]["UserID"], temp.ID);
            Assert.AreEqual(userTable.Rows[0]["Name"], temp.Name);
            Assert.AreEqual(userTable.Rows[0]["Status"], (int)temp.Status);
            ticks = DateTime.Now.Ticks;
            User[] users2 = new User[userTable.Rows.Count];
            for  (int i = 0; i < userTable.Rows.Count; ++i)
            {
                users2[i] = (User)mapper.ConvertObject(userTable.Rows[i], new User());
            }
            long ticks2 = DateTime.Now.Ticks - ticks;
            Console.WriteLine(string.Format("{0} ({1})", ticks2, ticks2*1.0/ticks1));
            Console.WriteLine(" vs ");

            //reflection
            ticks = DateTime.Now.Ticks;
            User[] users3 = new User[userTable.Rows.Count];
            for  (int i = 0; i < userTable.Rows.Count; ++i)
            {
                User newUser = new User();
                typeof(User).GetProperty("ID").SetValue(newUser, userTable.Rows[i]["UserID"], null);
                typeof(User).GetField("Name").SetValue(newUser, userTable.Rows[i]["Name"]);
                typeof(User).GetField("Status").SetValue(newUser, (UserStatus)userTable.Rows[i]["Status"]);
                users3[i] = newUser;
            }
            long ticks3 = DateTime.Now.Ticks - ticks;
            Console.WriteLine(string.Format("{0} ({1})", ticks3, ticks3*1.0/ticks1));
        }
    }

    public class User
    {
        private int? id;

        public int? ID
        {
            get
            {
                return id;
            }
            set
            {
                id = value;
            }
        }
        
        public string Name;

        public UserStatus Status;

        public User()
        {
        }

        public User(int? id, string name, UserStatus status)
        {
            this.id = id;
            this.Name = name;
            this.Status = status;
        }
    }

    public enum UserStatus
    {
        Normal,
        Admin
    }
}
