using System;
using System.Collections.Generic;
using System.Reflection;

namespace NBearMapping
{
    internal static class CustomObjectMemberHandlerCache
    {
        internal delegate List<ObjectMember> CustomObjectMemberGetHandler(object targetObject);
        internal delegate void CustomObjectMemberSetHandler(object targetObject, Dictionary<string, ObjectMember> members);

        private static Dictionary<Type, CustomObjectMemberGetHandler>  cachedMemberGetHandlers = new Dictionary<Type, CustomObjectMemberGetHandler>();
        private static Dictionary<Type, CustomObjectMemberSetHandler>  cachedMemberSetHandlers = new Dictionary<Type, CustomObjectMemberSetHandler>();

        internal static CustomObjectMemberGetHandler GetMemberGetter(Type objType)
        {
            if (!cachedMemberGetHandlers.ContainsKey(objType))
            {
                lock (cachedMemberGetHandlers)
                {
                    if (!cachedMemberGetHandlers.ContainsKey(objType))
                    {
                        Emit.CodeGenerator gen = new Emit.CodeGenerator();


                        gen.BeginMethod("m" + Guid.NewGuid().ToString("N"), typeof(CustomObjectMemberGetHandler));
                        
                        //init local variable: members
                        gen.DeclareLocal(typeof(List<ObjectMember>), "members");
                        gen.New(typeof(List<ObjectMember>).GetConstructor(Type.EmptyTypes));
                        gen.Stloc(0);

                        List<string> memberNames = new List<string>();
                        Type type = objType;
                        while (type != null && type != typeof(Object) && type != typeof(ValueType))
                        {
                            foreach (FieldInfo fi in type.GetFields())
                            {
                                if ((!fi.IsInitOnly) && (!memberNames.Contains(fi.Name)))
                                {
                                    gen.Ldloc(0);

                                    //name
                                    gen.Ldstr(fi.Name);

                                    //type
                                    gen.Ldtoken(CommonUtils.GetOriginalTypeOfNullableType(fi.FieldType.IsEnum ? typeof(int) : fi.FieldType));
                                    gen.Call(typeof(Type).GetMethod("GetTypeFromHandle", new Type[] { typeof(RuntimeTypeHandle) }));

                                    //value
                                    gen.Ldarg(0);
                                    gen.LoadMember(fi);
                                    if (fi.FieldType.IsValueType)
                                    {
                                        gen.Box(fi.FieldType.IsEnum ? typeof(int) : fi.FieldType);
                                    }

                                    //new ObjectMember
                                    gen.New(typeof(ObjectMember).GetConstructor(new Type[] { typeof(string), typeof(Type), typeof(object) }));

                                    //add to members
                                    gen.Call(typeof(List<ObjectMember>).GetMethod("Add"));

                                    memberNames.Add(fi.Name);
                                }
                            }

                            foreach (PropertyInfo pi in type.GetProperties())
                            {
                                if (pi.CanRead && pi.CanWrite &&(!memberNames.Contains(pi.Name)))
                                {
                                    gen.Ldloc(0);

                                    //name
                                    gen.Ldstr(pi.Name);

                                    //type
                                    gen.Ldtoken(CommonUtils.GetOriginalTypeOfNullableType(pi.PropertyType.IsEnum ? typeof(int) : pi.PropertyType));
                                    gen.Call(typeof(Type).GetMethod("GetTypeFromHandle", new Type[] { typeof(RuntimeTypeHandle) }));

                                    //value
                                    gen.Ldarg(0);
                                    gen.LoadMember(pi);
                                    if (pi.PropertyType.IsValueType)
                                    {
                                        gen.Box(pi.PropertyType.IsEnum ? typeof(int) : pi.PropertyType);
                                    }

                                    //new ObjectMember
                                    gen.New(typeof(ObjectMember).GetConstructor(new Type[] { typeof(string), typeof(Type), typeof(object) }));

                                    //add to members
                                    gen.Call(typeof(List<ObjectMember>).GetMethod("Add"));

                                    memberNames.Add(pi.Name);
                                }
                            }

                            type = type.BaseType;
                        }

                        gen.Ldloc(0);

                        cachedMemberGetHandlers.Add(objType, (CustomObjectMemberGetHandler)gen.EndMethod());
                    }
                }
            }

            return cachedMemberGetHandlers[objType];
        }

        internal static CustomObjectMemberSetHandler GetMemberSetter(Type objType)
        {
            if (!cachedMemberSetHandlers.ContainsKey(objType))
            {
                lock (cachedMemberSetHandlers)
                {
                    if (!cachedMemberSetHandlers.ContainsKey(objType))
                    {
                        Emit.CodeGenerator gen = new Emit.CodeGenerator();
                        gen.BeginMethod("m" + Guid.NewGuid().ToString("N"), typeof(CustomObjectMemberSetHandler));
                        
                        List<string> memberNames = new List<string>();
                        Type type = objType;
                        while (type != null && type != typeof(Object) && type != typeof(ValueType))
                        {
                            foreach (FieldInfo fi in type.GetFields())
                            {
                                if ((!fi.IsInitOnly) && (!memberNames.Contains(fi.Name)))
                                {
                                    gen.Ldarg(1);
                                    gen.Ldstr(fi.Name);
                                    gen.Call(typeof(Dictionary<string, ObjectMember>).GetMethod("ContainsKey"));
                                    gen.Load(true);
                                    gen.If(Emit.Cmp.EqualTo);
                                    gen.Ldtoken(CommonUtils.GetOriginalTypeOfNullableType(fi.FieldType.IsEnum ? typeof(int) : fi.FieldType));
                                    gen.Call(typeof(Type).GetMethod("GetTypeFromHandle", new Type[] { typeof(RuntimeTypeHandle) }));
                                    gen.Ldarg(1);
                                    gen.Ldstr(fi.Name);
                                    gen.Call(typeof(Dictionary<string, ObjectMember>).GetMethod("get_Item"));
                                    gen.LoadMember(typeof(ObjectMember).GetField("Type"));
                                    gen.If(Emit.Cmp.EqualTo);
                                    gen.Ldarg(0);
                                    gen.Ldarg(1);
                                    gen.Ldstr(fi.Name);
                                    gen.Call(typeof(Dictionary<string, ObjectMember>).GetMethod("get_Item"));
                                    gen.LoadMember(typeof(ObjectMember).GetField("Value"));
                                    gen.Load(null);
                                    gen.If(Emit.Cmp.EqualTo);
                                    if (!fi.FieldType.ToString().StartsWith("System.Nullable`1["))
                                    {
                                        gen.LoadDefaultValue(fi.FieldType);
                                    }
                                    else
                                    {
                                        gen.Load(null);
                                    }
                                    gen.Else();
                                    gen.Ldarg(1);
                                    gen.Ldstr(fi.Name);
                                    gen.Call(typeof(Dictionary<string, ObjectMember>).GetMethod("get_Item"));
                                    gen.LoadMember(typeof(ObjectMember).GetField("Value"));
                                    if (fi.FieldType.IsValueType)
                                    {
                                        gen.UnboxAny(fi.FieldType.IsEnum ? typeof(int) : fi.FieldType);
                                    }
                                    gen.EndIf();
                                    gen.StoreMember(fi);
                                    gen.EndIf();
                                    gen.EndIf();
                                    
                                    memberNames.Add(fi.Name);
                                }
                            }

                            foreach (PropertyInfo pi in type.GetProperties())
                            {
                                if (pi.CanRead && pi.CanWrite &&(!memberNames.Contains(pi.Name)))
                                {
                                    gen.Ldarg(1);
                                    gen.Ldstr(pi.Name);
                                    gen.Call(typeof(Dictionary<string, ObjectMember>).GetMethod("ContainsKey"));
                                    gen.Load(true);
                                    gen.If(Emit.Cmp.EqualTo);
                                    gen.Ldtoken(CommonUtils.GetOriginalTypeOfNullableType(pi.PropertyType.IsEnum ? typeof(int) : pi.PropertyType));
                                    gen.Call(typeof(Type).GetMethod("GetTypeFromHandle", new Type[] { typeof(RuntimeTypeHandle) }));
                                    gen.Ldarg(1);
                                    gen.Ldstr(pi.Name);
                                    gen.Call(typeof(Dictionary<string, ObjectMember>).GetMethod("get_Item"));
                                    gen.LoadMember(typeof(ObjectMember).GetField("Type"));
                                    gen.If(Emit.Cmp.EqualTo);
                                    gen.Ldarg(0);
                                    gen.Ldarg(1);
                                    gen.Ldstr(pi.Name);
                                    gen.Call(typeof(Dictionary<string, ObjectMember>).GetMethod("get_Item"));
                                    gen.LoadMember(typeof(ObjectMember).GetField("Value"));
                                    gen.Load(null);
                                    gen.If(Emit.Cmp.EqualTo);
                                    if (!pi.PropertyType.ToString().StartsWith("System.Nullable`1["))
                                    {
                                        gen.LoadDefaultValue(pi.PropertyType);
                                    }
                                    else
                                    {
                                        gen.Load(null);
                                    }
                                    gen.Else();
                                    gen.Ldarg(1);
                                    gen.Ldstr(pi.Name);
                                    gen.Call(typeof(Dictionary<string, ObjectMember>).GetMethod("get_Item"));
                                    gen.LoadMember(typeof(ObjectMember).GetField("Value"));
                                    if (pi.PropertyType.IsValueType)
                                    {
                                        gen.UnboxAny(pi.PropertyType.IsEnum ? typeof(int) : pi.PropertyType);
                                    }
                                    gen.EndIf();
                                    gen.StoreMember(pi);
                                    gen.EndIf();
                                    gen.EndIf();

                                    memberNames.Add(pi.Name);
                                }
                            }

                            type = type.BaseType;
                        }

                        cachedMemberSetHandlers.Add(objType, (CustomObjectMemberSetHandler)gen.EndMethod());
                    }
                }
            }

            return cachedMemberSetHandlers[objType];
        }
    }
}
