using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Data;

namespace NBearMapping
{
    public interface IMemberGetter
    {
        List<ObjectMember> GetMembers(object targetObj);
    }

    public static class MemberGetterFactory
    {
        private class MemberGetterInfo
        {
            public bool AsSingleton = true;
            public Type MemberGetterType;
            public object SigletonInstance;

            public MemberGetterInfo(bool asSingleton, Type type, object singletonInstance)
            {
                this.AsSingleton = asSingleton;
                this.MemberGetterType = type;
                this.SigletonInstance = singletonInstance;
            }
        }

        private static Dictionary<Type, MemberGetterInfo> memberGetters = new Dictionary<Type, MemberGetterInfo>();

        public static IMemberGetter GetMemberGetter(Type objType)
        {
            Check.Require(objType != null, "objType could not be null.");

            if (memberGetters.ContainsKey(objType))
            {
                MemberGetterInfo info = memberGetters[objType];
                return (IMemberGetter)(info.AsSingleton ? info.SigletonInstance : Activator.CreateInstance(info.MemberGetterType));
            }

            if (typeof(IDataReader).IsAssignableFrom(objType))
            {
                return DataReaderMemberGetter.Instance;
            }
            else if (typeof(DataRow).IsAssignableFrom(objType))
            {
                return DataRowMemberGetter.Instance;
            }
            else //use custom object member getter
            {
                return new CustomObjectMemberGetter(objType);
            }
        }

        public static void RegisterMemberGetter(Type objType, Type memberGetterType, bool asSingleton)
        {
            Check.Require(objType != null, "objType could not be null.");
            Check.Require(memberGetterType != null, "memberGetterType could not be null.");
            Check.Require(memberGetterType.GetConstructor(Type.EmptyTypes) != null, "memberGetterType MUST have default constructor.");

            lock (memberGetters)
            {
                if (memberGetters.ContainsKey(objType))
                {
                    memberGetters[objType] = new MemberGetterInfo(asSingleton, memberGetterType, asSingleton ? Activator.CreateInstance(memberGetterType) : null);
                }
                else
                {
                    memberGetters.Add(objType, new MemberGetterInfo(asSingleton, memberGetterType, asSingleton ? Activator.CreateInstance(memberGetterType) : null));
                }
            }
        }

        public static void RegisterMemberGetter(Type objType, object memberGetterInstance)
        {
            Check.Require(objType != null, "objType could not be null.");
            Check.Require(memberGetterInstance != null, "memberGetterInstance could not be null.");

            lock (memberGetters)
            {
                if (memberGetters.ContainsKey(objType))
                {
                    memberGetters[objType] = new MemberGetterInfo(true, memberGetterInstance.GetType(), memberGetterInstance);
                }
                else
                {
                    memberGetters.Add(objType, new MemberGetterInfo(true, memberGetterInstance.GetType(), memberGetterInstance));
                }
            }
        }

        public static void RemoveMemberGetter(Type objType)
        {
            lock (memberGetters)
            {
                if (memberGetters.ContainsKey(objType))
                    memberGetters.Remove(objType);
            }
        }

    }
}
