using System;
using System.Collections.Generic;
using System.Reflection;
using System.Data;

namespace NBearMapping
{
    public interface IMemberSetter
    {
        void SetMembers(object targetObj, Dictionary<string, ObjectMember> members);
    }

    public class MemberSetterFactory
    {
        private class MemberSetterInfo
        {
            public bool AsSingleton = true;
            public Type MemberSetterType;
            public object SigletonInstance;

            public MemberSetterInfo(bool asSingleton, Type type, object singletonInstance)
            {
                this.AsSingleton = asSingleton;
                this.MemberSetterType = type;
                this.SigletonInstance = singletonInstance;
            }
        }

        private static Dictionary<Type, MemberSetterInfo> memberSetters = new Dictionary<Type, MemberSetterInfo>();

        public static IMemberSetter GetMemberSetter(Type objType)
        {
            Check.Require(objType != null, "objType could not be null.");

            if (memberSetters.ContainsKey(objType))
            {
                MemberSetterInfo info = memberSetters[objType];
                return (IMemberSetter)(info.AsSingleton ? info.SigletonInstance : Activator.CreateInstance(info.MemberSetterType));
            }

            if (typeof(DataRow).IsAssignableFrom(objType))
            {
                return DataRowMemberSetter.Instance;
            }
            else //use custom object member getter
            {
                return new CustomObjectMemberSetter(objType);
            }
        }

        public static void RegisterMemberSetter(Type objType, Type memberSetterType, bool asSingleton)
        {
            Check.Require(objType != null, "objType could not be null.");
            Check.Require(memberSetterType != null, "memberSetterType could not be null.");
            Check.Require(memberSetterType.GetConstructor(Type.EmptyTypes) != null, "memberSetterType MUST have default constructor.");

            lock (memberSetters)
            {
                if (memberSetters.ContainsKey(objType))
                {
                    memberSetters[objType] = new MemberSetterInfo(asSingleton, memberSetterType, asSingleton ? Activator.CreateInstance(memberSetterType) : null);
                }
                else
                {
                    memberSetters.Add(objType, new MemberSetterInfo(asSingleton, memberSetterType, asSingleton ? Activator.CreateInstance(memberSetterType) : null));
                }
            }
        }

        public static void RegisterMemberSetter(Type objType, object memberSetterInstance)
        {
            Check.Require(objType != null, "objType could not be null.");
            Check.Require(memberSetterInstance != null, "memberSetterInstance could not be null.");

            lock (memberSetters)
            {
                if (memberSetters.ContainsKey(objType))
                {
                    memberSetters[objType] = new MemberSetterInfo(true, memberSetterInstance.GetType(), memberSetterInstance);
                }
                else
                {
                    memberSetters.Add(objType, new MemberSetterInfo(true, memberSetterInstance.GetType(), memberSetterInstance));
                }
            }
        }

        public static void RemoveMemberSetter(Type objType)
        {
            lock (memberSetters)
            {
                if (memberSetters.ContainsKey(objType))
                    memberSetters.Remove(objType);
            }
        }
    }
}
