using System;
using System.Collections.Generic;
using System.Text;
using System.Data;

namespace NBearMapping
{
    public delegate void CustomObjectMemberMappingHandler(object inputObject, object outputObject);

    public class ObjectMapper
    {
        #region Private Members

        private Type inputType;
        private Type outputType;
        private Type realOutputType;
        private Dictionary<string, string> mappingNames = new Dictionary<string, string>();

        private DataTable cachedTable;

        private DataTable ConstructDataTableSchema(string tableName, Dictionary<string, ObjectMember> members)
        {
            if (cachedTable != null)
                return (DataTable)cachedTable.Clone();

            Check.Require(!string.IsNullOrEmpty(tableName), "tableName could not be null or empty.");
            Check.Require(members != null, "members could not be null.");

            DataTable table = new DataTable(tableName);

            Dictionary<string, ObjectMember>.Enumerator en = members.GetEnumerator();
            while (en.MoveNext())
            {
                DataColumn column = new DataColumn(en.Current.Key, en.Current.Value.Type);
                table.Columns.Add(column);
            }

            cachedTable = table;
            return table;
        }

        #endregion

        public ObjectMapper(Type inputType, Type outputType)
        {
            Check.Require(inputType != null, "inputType could not be null.");
            Check.Require(outputType != null, "outputType could not be null.");
            Check.Require(!typeof(IDataReader).IsAssignableFrom(outputType), "outputType could not be a DataReader.");

            this.inputType = inputType;
            this.realOutputType = outputType;
            if (typeof(IDataReader) == outputType || typeof(DataRow) == outputType || typeof(DataTable) == outputType)
            {
                this.outputType = typeof(DataRow);
            }
            else
            {
                this.outputType = outputType;
            }
        }

        public void AddCustomMappingName(string inputMemberName, string outputMemberName)
        {
            Check.Require(!string.IsNullOrEmpty(inputMemberName), "inputMemberName could not be null or empty.");
            Check.Require(!string.IsNullOrEmpty(outputMemberName), "outputMemberName could not be null or empty.");

            lock (mappingNames)
            {
                if (mappingNames.ContainsKey(inputMemberName))
                {
                    mappingNames[inputMemberName] = outputMemberName;
                }
                else
                {
                    mappingNames.Add(inputMemberName, outputMemberName);
                }
            }
        }

        public void ClearCustomMappingName()
        {
            lock (mappingNames)
            {
                mappingNames.Clear();
            }
        }

        public void RemoveCustomMappingName(string inputMemberName)
        {
            Check.Require(!string.IsNullOrEmpty(inputMemberName), "inputMemberName could not be null or empty.");

            lock (mappingNames)
            {
                mappingNames.Remove(inputMemberName);
            }
        }

        /// <summary>
        /// Convert the specified inputObject to the output mapping type.
        /// </summary>
        /// <param name="inputObject">The inputObject must be consistent with the specified inputMappingType in ObjectMapper's Constructor.</param>
        /// <returns></returns>
        public object ConvertObject(object inputObject, params CustomObjectMemberMappingHandler[] customMappingHandlers)
        {
            Check.Require(inputObject != null, "inputObject could not be null.");
            Check.Require(outputType.GetConstructor(Type.EmptyTypes) != null, "outputType MUST have default constructor.");

            return ConvertObject(inputObject, null, customMappingHandlers);
        }        

        /// <summary>
        /// Convert the specified inputObject to the output mapping type.
        /// </summary>
        /// <param name="inputObject">The inputObject must be consistent with the specified inputMappingType in ObjectMapper's Constructor.</param>
        /// <param name="outputObject">outputObject is an instance to set member values on instead of create a new one.</param>
        /// <returns></returns>
        public object ConvertObject(object inputObject, object outputObject, params CustomObjectMemberMappingHandler[] customMappingHandlers)
        {
            Check.Require(inputObject != null, "inputObject could not be null.");

            IMemberGetter getter = MemberGetterFactory.GetMemberGetter(inputType);
            IMemberSetter setter = MemberSetterFactory.GetMemberSetter(outputType);
            List<ObjectMember> getMembers = getter.GetMembers(inputObject);
            Dictionary<string, ObjectMember> setMembers = new Dictionary<string, ObjectMember>();
            for (int i = 0; i < getMembers.Count; ++i)
            {
                setMembers.Add(mappingNames.ContainsKey(getMembers[i].Name) ? mappingNames[getMembers[i].Name] : getMembers[i].Name, getMembers[i]);
            }

            
            if (typeof(DataRow) == outputType)
            {
                if (outputObject == null)
                    outputObject = ConstructDataTableSchema(inputType.Name, setMembers);
                if (outputObject is DataTable)
                {
                    DataTable table = outputObject as DataTable;
                    DataRow row = table.NewRow();
                    setter.SetMembers(row, setMembers);
                    table.Rows.Add(row);
                    table.AcceptChanges();
                }
                else
                {
                    setter.SetMembers(outputObject, setMembers);
                }
            }
            else
            {
                if (outputObject == null)
                    outputObject = Activator.CreateInstance(outputType);
                setter.SetMembers(outputObject, setMembers);
            }

            if (realOutputType == typeof(IDataReader))
            {
                outputObject = (outputObject as DataTable).CreateDataReader();
            }
            else if (realOutputType == typeof(DataRow))
            {
                outputObject = (outputObject as DataTable).Rows[(outputObject as DataTable).Rows.Count - 1];
            }

            if (customMappingHandlers != null && customMappingHandlers.Length > 0)
            {
                for (int i = 0; i < customMappingHandlers.Length; ++i)
                {
                    if (customMappingHandlers[i] != null)
                        customMappingHandlers[i](inputObject, outputObject);
                }
            }

            return outputObject;
        }
    }
}
