using System;
using System.Collections.Generic;
using System.Reflection;

namespace NBearMapping
{
    public sealed class ReflectionUtils
    {
        private ReflectionUtils() { }

        public static PropertyInfo[] DeepGetProperties(params Type[] types)
        {
            if (types == null || types.Length == 0)
            {
                return new PropertyInfo[0];
            }
            List<PropertyInfo> list = new List<PropertyInfo>();
            foreach (Type t in types)
            {
                if (t != null)
                {
                    foreach (PropertyInfo pi in t.GetProperties())
                    {
                        list.Add(pi);
                    }

                    if (t.IsInterface)
                    {
                        Type[] interfaceTypes = t.GetInterfaces();

                        if (interfaceTypes != null)
                        {
                            foreach (PropertyInfo pi in DeepGetProperties(interfaceTypes))
                            {
                                bool isContained = false;

                                foreach (PropertyInfo item in list)
                                {
                                    if (item.Name == pi.Name)
                                    {
                                        isContained = true;
                                        break;
                                    }
                                }

                                if (!isContained)
                                {
                                    list.Add(pi);
                                }
                            }
                        }
                    }
                    else
                    {
                        Type baseType = t.BaseType;

                        if (baseType != typeof(object) && baseType != typeof(ValueType))
                        {
                            foreach (PropertyInfo pi in DeepGetProperties(baseType))
                            {
                                bool isContained = false;

                                foreach (PropertyInfo item in list)
                                {
                                    if (item.Name == pi.Name)
                                    {
                                        isContained = true;
                                        break;
                                    }
                                }

                                if (!isContained)
                                {
                                    list.Add(pi);
                                }
                            }
                        }
                    }
                }
            }

            return list.ToArray();
        }

        public static FieldInfo[] DeepGetFields(params Type[] types)
        {
            if (types == null || types.Length == 0)
            {
                return new FieldInfo[0];
            }
            List<FieldInfo> list = new List<FieldInfo>();
            foreach (Type t in types)
            {
                if (t != null)
                {
                    foreach (FieldInfo fi in t.GetFields())
                    {
                        list.Add(fi);
                    }

                    if (t.IsInterface)
                    {
                        Type[] interfaceTypes = t.GetInterfaces();

                        if (interfaceTypes != null)
                        {
                            foreach (FieldInfo fi in DeepGetFields(interfaceTypes))
                            {
                                bool isContained = false;

                                foreach (FieldInfo item in list)
                                {
                                    if (item.Name == fi.Name)
                                    {
                                        isContained = true;
                                        break;
                                    }
                                }

                                if (!isContained)
                                {
                                    list.Add(fi);
                                }
                            }
                        }
                    }
                    else
                    {
                        Type baseType = t.BaseType;

                        if (baseType != typeof(object) && baseType != typeof(ValueType))
                        {
                            foreach (FieldInfo fi in DeepGetFields(baseType))
                            {
                                bool isContained = false;

                                foreach (FieldInfo item in list)
                                {
                                    if (item.Name == fi.Name)
                                    {
                                        isContained = true;
                                        break;
                                    }
                                }

                                if (!isContained)
                                {
                                    list.Add(fi);
                                }
                            }
                        }
                    }
                }
            }

            return list.ToArray();
        }

        /// <summary>
        /// Deeply get property info from specified type.
        /// </summary>
        /// <param name="type">The type.</param>
        /// <param name="propertyName">Name of the property.</param>
        /// <returns></returns>
        public static PropertyInfo DeepGetProperty(Type type, string propertyName)
        {
            foreach (PropertyInfo pi in DeepGetProperties(type))
            {
                if (pi.Name == propertyName)
                {
                    return pi;
                }
            }

            return null;
        }

        /// <summary>
        /// Deeps the get field from specific type.
        /// </summary>
        /// <param name="type">The type.</param>
        /// <param name="name">The name.</param>
        /// <param name="isPublic">if is public.</param>
        /// <returns>The field info</returns>
        public static FieldInfo DeepGetField(Type type, string name, bool includePublic, bool includeNonPublic, bool isStatic)
        {
            Type t = type;
            if (t != null)
            {
                BindingFlags flags = BindingFlags.Instance;
                if (includePublic)
                    flags |= BindingFlags.Public;
                if (includeNonPublic)
                    flags |= BindingFlags.NonPublic;
                if (isStatic)
                    flags |= BindingFlags.Static;
                FieldInfo fi = t.GetField(name, flags);
                if (fi != null)
                {
                    return fi;
                }

                if (t.IsInterface)
                {
                    Type[] interfaceTypes = t.GetInterfaces();

                    if (interfaceTypes != null)
                    {
                        foreach (Type interfaceType in interfaceTypes)
                        {
                            fi = DeepGetField(interfaceType, name, includePublic, includeNonPublic, isStatic);
                            if (fi != null)
                            {
                                return fi;
                            }
                        }
                    }
                }
                else
                {
                    Type baseType = t.BaseType;

                    if (baseType != typeof(object) && baseType != typeof(ValueType))
                    {
                        return DeepGetField(baseType, name, includePublic, includeNonPublic, isStatic);
                    }
                }
            }
            return null;
        }

        public static MethodInfo GetMethodInfoFromArrayBySignature(string signature, MethodInfo[] mis)
        {
            if (mis == null)
            {
                return null;
            }

            foreach (MethodInfo mi in mis)
            {
                if (mi.ToString() == signature)
                {
                    return mi;
                }
            }

            return null;
        }

        public static MethodInfo GetMethodInfoBySignature(Type type, string signature, bool includePublic, bool includeNonPublic, bool isStatic)
        {
            if (type == null)
            {
                throw new ArgumentNullException("type");
            }

            BindingFlags flags = BindingFlags.Instance;
            if (includePublic)
                flags |= BindingFlags.Public;
            if (includeNonPublic)
                flags |= BindingFlags.NonPublic;
            if (isStatic)
                flags |= BindingFlags.Static;
            return GetMethodInfoFromArrayBySignature(signature, type.GetMethods(flags));
        }
    }
}
